"""
    savegif(filename::String, o::Output; kw...)

Write the output array to a gif.

# Arguments

- `filename`: File path to save the gif file to.
- `output`: An [`Output`](@ref) object. Note that to make a gif, the output should stores
    frames, and run with `store=true`, and `@assert DynamicGrids.istored(o)` should pass.

# Keywords

[`ImageConfig`](@ref) keywords:

$IMAGECONFIG_KEYWORDS
"""
function savegif(filename::String, o::Output, ruleset=Ruleset();
    minval=minval(o), maxval=maxval(o),
    scheme=ObjectScheme(), zerocolor=ZEROCOL, maskcolor=MASKCOL,
    renderer=autorenderer(init(o); scheme, zerocolor, maskcolor),
    font=autofont(), text=TextConfig(font=font), textconfig=text, kw...
)
    im_o = NoDisplayImageOutput(o;
        imageconfig=ImageConfig(init(o);
            minval=minval, maxval=maxval, renderer=renderer, textconfig=textconfig
        )
    )
    savegif(filename, im_o, ruleset; kw...)
end
function savegif(filename::String, o::ImageOutput, ruleset=Ruleset(); fps=fps(o), kw...)
    length(o) == 1 && @warn "The output has length 1: the saved gif will be a single image"
    data = SimData(o, ruleset)
    imsize = size(first(grids(data)))
    gif = Array{ARGB32}(undef, imsize..., length(o))
    foreach(firstindex(o):lastindex(o)) do f
        @set! data.currentframe = f
        render!(view(gif, :, :, f), renderer(o), o, data, o[f])
    end
    FileIO.save(File{format"GIF"}(filename), gif; fps=fps, kw...)
    return filename
end

"""
    GifOutput <: ImageOutput

    GifOutput(init; filename, tspan, kw...)

Output that stores the simulation as images and saves a Gif file on completion.

# Arguments:

- `init`: initialisation `AbstractArrayArray` or `NamedTuple` of `AbstractArrayArray`.

# Keywords

Storing the gif:
- `filename`: File path to save the gif file to.

$IMAGEOUTPUT_KEYWORDS
"""
mutable struct GifOutput{T,F<:AbstractVector{T},E,GC,IC,G,N} <: ImageOutput{T,F}
    frames::F
    running::Bool
    extent::E
    graphicconfig::GC
    imageconfig::IC
    gif::G
    filename::N
end
function GifOutput(; frames, running, extent, graphicconfig, imageconfig, filename, kw...)
    gif = _allocgif(imageconfig, extent)
    GifOutput(frames, running, extent, graphicconfig, imageconfig, gif, filename)
end

# Getters
filename(o::GifOutput) = o.filename
gif(o::GifOutput) = o.gif

# Output/ImageOutput interface methods
maybesleep(output::GifOutput, f) = nothing
function showimage(image, o::GifOutput, data::AbstractSimData)
    gif(o)[:, :, currentframe(data)] .= image
end
finalisegraphics(o::GifOutput, data::AbstractSimData) = savegif(o)

# The gif is already generated, just save it again if neccessary
savegif(o::GifOutput) = savegif(filename(o), o)
function savegif(filename::String, o::GifOutput, fps=fps(o); kw...)
    FileIO.save(File{format"GIF"}(filename), gif(o); fps=fps, kw...)
end

# Preallocate the 3 dimensional gif array
_allocgif(i::ImageConfig, e::Extent) = _allocgif(renderer(i), i::ImageConfig, e::Extent)
function _allocgif(r::Renderer, i::ImageConfig, e::Extent)
    zeros(ARGB32, imagesize(r, e)..., length(tspan(e)))
end
